# === sendOrderMsg_Sys ===
import datetime
from abc import ABC

from pymammotion import logger
from pymammotion.mammotion.commands.abstract_message import AbstractMessage
from pymammotion.mammotion.commands.messages.navigation import MessageNavigation
from pymammotion.proto.luba_msg import LubaMsg, MsgAttr, MsgCmdType, MsgDevice
from pymammotion.proto.mctrl_sys import (
    DeviceProductTypeInfoT,
    LoraCfgReq,
    MctlSys,
    MCtrlSimulationCmdData,
    ReportInfoCfg,
    RptAct,
    RptInfoType,
    SysCommCmd,
    SysKnifeControl,
    SysSetDateTime,
    TimeCtrlLight,
)
from pymammotion.utility.device_type import DeviceType


class MessageSystem(AbstractMessage, ABC):
    messageNavigation: MessageNavigation = MessageNavigation()

    def send_order_msg_sys(self, sys):
        luba_msg = LubaMsg(
            msgtype=MsgCmdType.MSG_CMD_TYPE_EMBED_SYS,
            sender=MsgDevice.DEV_MOBILEAPP,
            rcver=MsgDevice.DEV_MAINCTL,
            sys=sys,
        )

        return luba_msg.SerializeToString()

    def reset_system(self):
        build = MctlSys(todev_reset_system=1)
        logger.debug("Send command - send factory reset")
        return self.send_order_msg_sys(build)

    def set_blade_control(self, on_off: int):
        mctlsys = MctlSys()
        sys_knife_control = SysKnifeControl()
        sys_knife_control.knife_status = on_off
        mctlsys.todev_knife_ctrl = sys_knife_control

        return self.send_order_msg_sys(mctlsys)

    def get_device_product_model(self):
        return self.send_order_msg_sys(MctlSys(device_product_type_info=DeviceProductTypeInfoT()))

    def read_and_set_sidelight(self, is_sidelight: bool, operate: int):
        """Read state of sidelight as well as set it."""
        if is_sidelight:
            build = TimeCtrlLight(
                operate=operate,
                enable=0,
                action=0,
                start_hour=0,
                start_min=0,
                end_hour=0,
                end_min=0,
            )
        else:
            build = TimeCtrlLight(
                operate=operate,
                enable=1,
                action=0,
                start_hour=0,
                start_min=0,
                end_hour=0,
                end_min=0,
            )
        logger.debug(f"Send read and write sidelight command is_sidelight:{
            is_sidelight}, operate:{operate}")
        build2 = MctlSys(todev_time_ctrl_light=build)
        logger.debug(f"Send command - send read and write sidelight command is_sidelight:{
            is_sidelight}, operate:{operate}, timeCtrlLight:{build}")
        return self.send_order_msg_sys(build2)

    def test_tool_order_to_sys(self, sub_cmd: int, param_id: int, param_value: list[int]):
        build = MCtrlSimulationCmdData(sub_cmd=sub_cmd, param_id=param_id, param_value=param_value)
        logger.debug(f"Send tool test command: subCmd={sub_cmd}, param_id:{
            param_id}, param_value={param_value}")
        build2 = MctlSys(simulation_cmd=build)
        logger.debug(f"Send tool test command: subCmd={sub_cmd}, param_id:{
            param_id}, param_value={param_value}")
        return self.send_order_msg_sys(build2)

    def read_and_set_rtk_paring_code(self, op: int, cgf: str | None = None):
        logger.debug(f"Send read and write base station configuration quality op:{
            op}, cgf:{cgf}")
        return self.send_order_msg_sys(MctlSys(todev_lora_cfg_req=LoraCfgReq(op=op, cfg=cgf)))

    def allpowerfull_rw(self, id: int, context: int, rw: int):
        if (id == 6 or id == 3 or id == 7) and DeviceType.is_luba_2(self.get_device_name()):
            self.messageNavigation.allpowerfull_rw_adapter_x3(id, context, rw)
            return
        build = MctlSys(bidire_comm_cmd=SysCommCmd(id=id, context=context, rw=rw))
        logger.debug(f"Send command - 9 general read and write command id={id}, context={context}, rw={rw}")
        if id == 5:
            # This logic doesnt make snese, but its what they had so..
            return self.send_order_msg_sys(build)
        return self.send_order_msg_sys(build)

    # Commented out as not needed and too many refs to try fix up
    # def factory_test_order(self, test_id: int, test_duration: int, expect: str):
    #     new_builder = mow_to_app_qctools_info_t.Builder()
    #     logger.debug(f"Factory tool logger.debug, expect={expect}")
    #     if not expect:
    #         build = new_builder.set_type_value(
    #             test_id).set_time_of_duration(test_duration).build()
    #     else:
    #         try:
    #             json_array = json.loads(expect)
    #             z2 = True
    #             for i in range(len(json_array)):
    #                 new_builder2 = QCAppTestExcept.Builder()
    #                 json_object = json_array[i]
    #                 if "except_type" in json_object:
    #                     string = json_object["except_type"]
    #                     if "conditions" in json_object:
    #                         json_array2 = json_object["conditions"]
    #                         for i2 in range(len(json_array2)):
    #                             json_object2 = json_array2[i2]
    #                             new_builder3 = QCAppTestConditions.Builder()
    #                             if "cond_type" in json_object2:
    #                                 new_builder3.set_cond_type(
    #                                     json_object2["cond_type"])
    #                             else:
    #                                 z2 = False
    #                             if "value" in json_object2:
    #                                 obj = json_object2["value"]
    #                                 if string == "int":
    #                                     new_builder3.set_int_val(int(obj))
    #                                 elif string == "float":
    #                                     new_builder3.set_float_val(float(obj))
    #                                 elif string == "double":
    #                                     new_builder3.set_double_val(float(obj))
    #                                 elif string == "string":
    #                                     new_builder3.set_string_val(str(obj))
    #                                 else:
    #                                     z2 = False
    #                                 new_builder2.add_conditions(new_builder3)
    #                             else:
    #                                 z2 = False
    #                     new_builder2.set_except_type(string)
    #                     new_builder.add_except(new_builder2)
    #                     new_builder2.clear()
    #             z = z2
    #         except json.JSONDecodeError:
    #             z = False
    #         if z:
    #             build = new_builder.set_type_value(
    #                 test_id).set_time_of_duration(test_duration).build()
    #         else:
    #             build = new_builder.set_type_value(
    #                 test_id).set_time_of_duration(test_duration).build()
    #     logger.debug(f"Factory tool logger.debug, mow_to_app_qctools_info_t={
    #         build.except_count}, mow_to_app_qctools_info_t22={build.except_list}")
    #     build2 = MctlSys(mow_to_app_qctools_info=build)
    #     logger.debug(f"Send command - factory tool test command testId={
    #         test_id}, testDuration={test_duration}", "Factory tool logger.debug222", True)
    #     return self.send_order_msg_sys(build2)

    def send_sys_set_date_time(self):
        calendar = datetime.datetime.now()
        i = calendar.year
        i2 = calendar.month
        i3 = calendar.day
        i4 = calendar.isoweekday()
        i5 = calendar.hour
        i6 = calendar.minute
        i7 = calendar.second
        i8 = calendar.utcoffset().total_seconds() // 60 if calendar.utcoffset() else 0
        i9 = 1 if calendar.dst() else 0
        logger.debug(f"Print time zone, time zone={
            i8}, daylight saving time={i9} week={i4}")
        build = MctlSys(
            todev_data_time=SysSetDateTime(
                year=i,
                month=i2,
                date=i3,
                week=i4,
                hours=i5,
                minutes=i6,
                seconds=i7,
                time_zone=i8,
                daylight=i9,
            )
        )
        logger.debug(
            f"Send command - synchronize time zone={i8}, daylight saving time={i9} week={i4}, day:{
            i3}, month:{i2}, hours:{i5}, minutes:{i6}, seconds:{i7}, year={i}",
            "Time synchronization",
            True,
        )
        return self.send_order_msg_sys(build)

    def get_device_version_info(self):
        return self.send_order_msg_sys(MctlSys(todev_get_dev_fw_info=1))

    # === sendOrderMsg_Sys2 ===

    def request_iot_sys(
        self,
        rpt_act: RptAct,
        rpt_info_type: list[RptInfoType | str] | None,
        timeout: int,
        period: int,
        no_change_period: int,
        count: int,
    ) -> bytes:
        build = MctlSys(
            todev_report_cfg=ReportInfoCfg(
                act=rpt_act,
                sub=rpt_info_type,
                timeout=timeout,
                period=period,
                no_change_period=no_change_period,
                count=count,
            )
        )
        logger.debug(f"Send command==== IOT slim data Act {
            build.todev_report_cfg.act} {build}")
        return self.send_order_msg_sys(build)

    def get_report_cfg_stop(self, timeout: int = 10000, period: int = 1000, no_change_period: int = 1000):
        mctlsys = MctlSys(
            todev_report_cfg=ReportInfoCfg(
                act=RptAct.RPT_STOP,
                timeout=timeout,
                period=period,
                no_change_period=no_change_period,
                count=1,
            )
        )

        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_CONNECT)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_RTK)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_DEV_LOCAL)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_WORK)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_DEV_STA)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_MAINTAIN)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VISION_POINT)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VIO)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VISION_STATISTIC)

        lubaMsg = LubaMsg()
        lubaMsg.msgtype = MsgCmdType.MSG_CMD_TYPE_EMBED_SYS
        lubaMsg.sender = MsgDevice.DEV_MOBILEAPP
        lubaMsg.rcver = MsgDevice.DEV_MAINCTL
        lubaMsg.msgattr = MsgAttr.MSG_ATTR_REQ
        lubaMsg.seqs = 1
        lubaMsg.version = 1
        lubaMsg.subtype = 1
        lubaMsg.sys = mctlsys
        return lubaMsg.SerializeToString()

    def get_report_cfg(self, timeout: int = 10000, period: int = 1000, no_change_period: int = 2000):
        mctlsys = MctlSys(
            todev_report_cfg=ReportInfoCfg(
                act=RptAct.RPT_START,
                timeout=timeout,
                period=period,
                no_change_period=no_change_period,
                count=1,
            )
        )

        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_CONNECT)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_RTK)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_DEV_LOCAL)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_WORK)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_DEV_STA)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_MAINTAIN)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VISION_POINT)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VIO)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_VISION_STATISTIC)
        mctlsys.todev_report_cfg.sub.append(RptInfoType.RIT_BASESTATION_INFO)

        lubaMsg = LubaMsg()
        lubaMsg.msgtype = MsgCmdType.MSG_CMD_TYPE_EMBED_SYS
        lubaMsg.sender = MsgDevice.DEV_MOBILEAPP
        lubaMsg.rcver = MsgDevice.DEV_MAINCTL
        lubaMsg.msgattr = MsgAttr.MSG_ATTR_REQ
        lubaMsg.seqs = 1
        lubaMsg.version = 1
        lubaMsg.subtype = 1
        lubaMsg.sys = mctlsys
        return lubaMsg.SerializeToString()



"""Device control of mammotion robots over bluetooth or MQTT."""

from __future__ import annotations

import asyncio
import logging
from enum import Enum
from functools import cache
from typing import Any

from bleak.backends.device import BLEDevice

from pymammotion.aliyun.cloud_gateway import CloudIOTGateway
from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model.account import Credentials
from pymammotion.data.model.device import MowingDevice
from pymammotion.http.http import connect_http
from pymammotion.mammotion.devices.mammotion_bluetooth import MammotionBaseBLEDevice
from pymammotion.mammotion.devices.mammotion_cloud import MammotionBaseCloudDevice, MammotionCloud
from pymammotion.mqtt import MammotionMQTT

TIMEOUT_CLOUD_RESPONSE = 10

_LOGGER = logging.getLogger(__name__)


class ConnectionPreference(Enum):
    """Enum for connection preference."""

    EITHER = 0
    WIFI = 1
    BLUETOOTH = 2


class MammotionMixedDeviceManager:
    _ble_device: MammotionBaseBLEDevice | None = None
    _cloud_device: MammotionBaseCloudDevice | None = None
    _mowing_state: MowingDevice
    preference: ConnectionPreference

    def __init__(
        self,
        name: str,
        cloud_device: Device | None = None,
        ble_device: BLEDevice | None = None,
        mqtt: MammotionCloud | None = None,
        preference: ConnectionPreference = ConnectionPreference.BLUETOOTH,
    ) -> None:
        self.name = name
        self._mowing_state = MowingDevice()
        self.add_ble(ble_device)
        self.add_cloud(cloud_device, mqtt)
        self.preference = preference

    def mower_state(self):
        return self._mowing_state

    def ble(self) -> MammotionBaseBLEDevice | None:
        return self._ble_device

    def cloud(self) -> MammotionBaseCloudDevice | None:
        return self._cloud_device

    def add_ble(self, ble_device: BLEDevice) -> None:
        if ble_device is not None:
            self._ble_device = MammotionBaseBLEDevice(self._mowing_state, ble_device)

    def add_cloud(self, cloud_device: Device | None = None, mqtt: MammotionCloud | None = None) -> None:
        if cloud_device is not None:
            self._cloud_device = MammotionBaseCloudDevice(
                mqtt, cloud_device=cloud_device, mowing_state=self._mowing_state
            )

    def replace_cloud(self, cloud_device: MammotionBaseCloudDevice) -> None:
        self._cloud_device = cloud_device

    def replace_ble(self, ble_device: MammotionBaseBLEDevice) -> None:
        self._ble_device = ble_device

    def replace_mqtt(self, mqtt: MammotionCloud) -> None:
        self._cloud_device._mqtt = mqtt

    def has_cloud(self) -> bool:
        return self._cloud_device is not None

    def has_ble(self) -> bool:
        return self._ble_device is not None


class MammotionDevices:
    devices: dict[str, MammotionMixedDeviceManager] = {}

    def add_device(self, mammotion_device: MammotionMixedDeviceManager) -> None:
        exists: MammotionMixedDeviceManager | None = self.devices.get(mammotion_device.name)
        if exists is None:
            self.devices[mammotion_device.name] = mammotion_device
            return
        if mammotion_device.has_cloud():
            exists.replace_cloud(mammotion_device.cloud())
        if mammotion_device.has_ble():
            exists.replace_ble(mammotion_device.ble())

    def get_device(self, mammotion_device_name: str) -> MammotionMixedDeviceManager:
        return self.devices.get(mammotion_device_name)

    async def remove_device(self, name) -> None:
        device_for_removal = self.devices.pop(name)
        if device_for_removal.has_cloud():
            should_disconnect = {
                device
                for key, device in self.devices.items()
                if device.cloud() is not None and device.cloud()._mqtt == device_for_removal.cloud()._mqtt
            }
            if len(should_disconnect) == 0:
                device_for_removal.cloud()._mqtt.disconnect()
            await device_for_removal.cloud().stop()
        if device_for_removal.has_ble():
            await device_for_removal.ble().stop()

        del device_for_removal


async def create_devices(
    ble_device: BLEDevice,
    cloud_credentials: Credentials | None = None,
    preference: ConnectionPreference = ConnectionPreference.BLUETOOTH,
):
    mammotion = Mammotion()
    mammotion.add_ble_device(ble_device, preference)

    if cloud_credentials and preference == ConnectionPreference.EITHER or preference == ConnectionPreference.WIFI:
        await mammotion.login_and_initiate_cloud(
            cloud_credentials.account_id or cloud_credentials.email, cloud_credentials.password
        )

    return mammotion


@cache
class Mammotion:
    """Represents a Mammotion account and its devices."""

    devices = MammotionDevices()
    mqtt_list: dict[str, MammotionCloud] = dict()

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        """Initialize MammotionDevice."""
        self._login_lock = asyncio.Lock()

    def add_ble_device(
        self, ble_device: BLEDevice, preference: ConnectionPreference = ConnectionPreference.BLUETOOTH
    ) -> None:
        if ble_device:
            self.devices.add_device(
                MammotionMixedDeviceManager(name=ble_device.name, ble_device=ble_device, preference=preference)
            )

    async def login_and_initiate_cloud(self, account, password, force: bool = False) -> None:
        async with self._login_lock:
            exists: MammotionCloud | None = self.mqtt_list.get(account)
            if not exists or force:
                cloud_client = await self.login(account, password)
                await self.initiate_cloud_connection(account, cloud_client)

    async def initiate_cloud_connection(self, account: str, cloud_client: CloudIOTGateway) -> None:
        if self.mqtt_list.get(account) is not None:
            if self.mqtt_list.get(account).is_connected:
                # we might have removed a device so readd
                self.add_cloud_devices(self.mqtt_list.get(account))
                return

        mammotion_cloud = MammotionCloud(
            MammotionMQTT(
                region_id=cloud_client.region_response.data.regionId,
                product_key=cloud_client.aep_response.data.productKey,
                device_name=cloud_client.aep_response.data.deviceName,
                device_secret=cloud_client.aep_response.data.deviceSecret,
                iot_token=cloud_client.session_by_authcode_response.data.iotToken,
                client_id=cloud_client.client_id,
                cloud_client=cloud_client,
            ),
            cloud_client,
        )
        self.mqtt_list[account] = mammotion_cloud
        self.add_cloud_devices(mammotion_cloud)

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.mqtt_list[account].connect_async)

    def add_cloud_devices(self, mqtt_client: MammotionCloud) -> None:
        for device in mqtt_client.cloud_client.devices_by_account_response.data.data:
            mower_device = self.devices.get_device(device.deviceName)
            if device.deviceName.startswith(("Luba-", "Yuka-")) and mower_device is None:
                self.devices.add_device(
                    MammotionMixedDeviceManager(
                        name=device.deviceName,
                        cloud_device=device,
                        mqtt=mqtt_client,
                        preference=ConnectionPreference.WIFI,
                    )
                )
            elif device.deviceName.startswith(("Luba-", "Yuka-")) and mower_device:
                if mower_device.cloud() is None:
                    mower_device.add_cloud(cloud_device=device, mqtt=mqtt_client)
                elif mqtt_client != mower_device.cloud().mqtt:
                    mower_device.replace_mqtt(mqtt_client)

    def set_disconnect_strategy(self, disconnect: bool) -> None:
        for device_name, device in self.devices.devices:
            if device.ble() is not None:
                ble_device: MammotionBaseBLEDevice = device.ble()
                ble_device.set_disconnect_strategy(disconnect)

    async def login(self, account: str, password: str) -> CloudIOTGateway:
        """Login to mammotion cloud."""
        cloud_client = CloudIOTGateway()
        mammotion_http = await connect_http(account, password)
        country_code = mammotion_http.login_info.userInformation.domainAbbreviation
        _LOGGER.debug("CountryCode: " + country_code)
        _LOGGER.debug("AuthCode: " + mammotion_http.login_info.authorization_code)
        cloud_client.set_http(mammotion_http)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None, cloud_client.get_region, country_code, mammotion_http.login_info.authorization_code
        )
        await cloud_client.connect()
        await cloud_client.login_by_oauth(country_code, mammotion_http.login_info.authorization_code)
        await loop.run_in_executor(None, cloud_client.aep_handle)
        await loop.run_in_executor(None, cloud_client.session_by_auth_code)

        await loop.run_in_executor(None, cloud_client.list_binding_by_account)
        return cloud_client

    def remove_device(self, name: str) -> None:
        self.devices.remove_device(name)

    def get_device_by_name(self, name: str) -> MammotionMixedDeviceManager:
        return self.devices.get_device(name)

    async def send_command(self, name: str, key: str):
        """Send a command to the device."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().command(key)
            if device.preference is ConnectionPreference.WIFI:
                return await device.cloud().command(key)
            # TODO work with both with EITHER

    async def send_command_with_args(self, name: str, key: str, **kwargs: Any):
        """Send a command with args to the device."""
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().command(key, **kwargs)
            if device.preference is ConnectionPreference.WIFI:
                return await device.cloud().command(key, **kwargs)
            # TODO work with both with EITHER

    async def start_sync(self, name: str, retry: int):
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().start_sync(retry)
            if device.preference is ConnectionPreference.WIFI:
                return await device.cloud().start_sync(retry)
            # TODO work with both with EITHER

    async def start_map_sync(self, name: str):
        device = self.get_device_by_name(name)
        if device:
            if device.preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().start_map_sync()
            if device.preference is ConnectionPreference.WIFI:
                return await device.cloud().start_map_sync()
            # TODO work with both with EITHER

    async def get_stream_subscription(self, name: str):
        device = self.get_device_by_name(name)
        if self._preference is ConnectionPreference.WIFI:
            if device.has_cloud():
                _stream_response = await self.cloud().cloud_client.get_stream_subscription(device.cloud().iot_id)
                _LOGGER.debug(_stream_response)
                return _stream_response

    def mower(self, name: str):
        device = self.get_device_by_name(name)
        if device:
            return device.mower_state()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymammotion/proto/mctrl_pept.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"pymammotion/proto/mctrl_pept.proto\"X\n\x16perception_obstacles_t\x12\r\n\x05label\x18\x01 \x01(\x05\x12\x0b\n\x03num\x18\x02 \x01(\x05\x12\x10\n\x08points_x\x18\x03 \x03(\x05\x12\x10\n\x08points_y\x18\x04 \x03(\x05\"\x98\x01\n$perception_obstacles_visualization_t\x12\x15\n\ris_heart_beat\x18\x01 \x01(\x05\x12\x0b\n\x03num\x18\x02 \x01(\x05\x12*\n\tobstacles\x18\x03 \x03(\x0b\x32\x17.perception_obstacles_t\x12\x11\n\ttimestamp\x18\x04 \x01(\x01\x12\r\n\x05scale\x18\x05 \x01(\x02\"f\n\x1bperception_universal_buff_t\x12\x17\n\x0fperception_type\x18\x01 \x01(\x05\x12\x16\n\x0eperception_len\x18\x02 \x01(\x05\x12\x16\n\x0euniversal_buff\x18\x03 \x03(\x03\"\xb0\x01\n\x08MctlPept\x12S\n\"perception_obstacles_visualization\x18\x01 \x01(\x0b\x32%.perception_obstacles_visualization_tH\x00\x12\x41\n\x19perception_universal_buff\x18\x02 \x01(\x0b\x32\x1c.perception_universal_buff_tH\x00\x42\x0c\n\nSubPeptMsgb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pymammotion.proto.mctrl_pept_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PERCEPTION_OBSTACLES_T._serialized_start=38
  _PERCEPTION_OBSTACLES_T._serialized_end=126
  _PERCEPTION_OBSTACLES_VISUALIZATION_T._serialized_start=129
  _PERCEPTION_OBSTACLES_VISUALIZATION_T._serialized_end=281
  _PERCEPTION_UNIVERSAL_BUFF_T._serialized_start=283
  _PERCEPTION_UNIVERSAL_BUFF_T._serialized_end=385
  _MCTLPEPT._serialized_start=388
  _MCTLPEPT._serialized_end=564
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass, field
from enum import IntEnum

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pymammotion.proto.mctrl_nav import AreaHashName, NavGetCommDataAck


class PathType(IntEnum):
    """Path types for common data."""

    AREA = 0
    OBSTACLE = 1
    PATH = 2


@dataclass
class FrameList(DataClassORJSONMixin):
    total_frame: int
    data: list[NavGetCommDataAck]


@dataclass
class AreaHashNameList(DataClassORJSONMixin):
    """Wrapper so we can serialize to and from dict."""

    name: str
    hash: int


@dataclass
class HashList(DataClassORJSONMixin):
    """stores our map data.
    [hashID, FrameList].
    hashlist for all our hashIDs for verification
    """

    area: dict = field(default_factory=dict)  # type 0
    path: dict = field(default_factory=dict)  # type 2
    obstacle: dict = field(default_factory=dict)  # type 1
    hashlist: list[int] = field(default_factory=list)
    area_name: list[AreaHashNameList] = field(default_factory=list)

    def set_hashlist(self, hashlist: list[int]) -> None:
        self.hashlist = hashlist
        self.area = {hash_id: frames for hash_id, frames in self.area.items() if hash_id in hashlist}
        self.path = {hash_id: frames for hash_id, frames in self.path.items() if hash_id in hashlist}
        self.obstacle = {hash_id: frames for hash_id, frames in self.obstacle.items() if hash_id in hashlist}

    def missing_frame(self, hash_data: NavGetCommDataAck) -> list[int]:
        if hash_data.type == PathType.AREA:
            return self._find_missing_frames(self.area.get(hash_data.hash))

        if hash_data.type == PathType.OBSTACLE:
            return self._find_missing_frames(self.obstacle.get(hash_data.hash))

        if hash_data.type == PathType.PATH:
            return self._find_missing_frames(self.path.get(hash_data.hash))

    def update(self, hash_data: NavGetCommDataAck) -> bool:
        """Update the map data."""
        if hash_data.type == PathType.AREA:
            existing_name = next((area for area in self.area_name if area.hash == hash_data.hash), None)
            if not existing_name:
                self.area_name.append(AreaHashName(name=f"area {len(self.area_name)+1}", hash=hash_data.hash))
            return self._add_hash_data(self.area, hash_data)

        if hash_data.type == PathType.OBSTACLE:
            return self._add_hash_data(self.obstacle, hash_data)

        if hash_data.type == PathType.PATH:
            return self._add_hash_data(self.path, hash_data)

    @staticmethod
    def _find_missing_frames(frame_list: FrameList) -> list[int]:
        if frame_list.total_frame == len(frame_list.data):
            return []
        number_list = list(range(1, frame_list.total_frame + 1))

        current_frames = {frame.current_frame for frame in frame_list.data}
        missing_numbers = [num for num in number_list if num not in current_frames]
        return missing_numbers

    @staticmethod
    def _add_hash_data(hash_dict: dict, hash_data: NavGetCommDataAck) -> bool:
        if hash_dict.get(hash_data.hash) is None:
            hash_dict[hash_data.hash] = FrameList(total_frame=hash_data.total_frame, data=[hash_data])
            return True

        if hash_data not in hash_dict[hash_data.hash].data:
            hash_dict[hash_data.hash].data.append(hash_data)
            return True
        return False



from dataclasses import dataclass
from typing import List

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class Camera(DataClassORJSONMixin):
    cameraId: int
    token: str


@dataclass
class StreamSubscriptionResponse(DataClassORJSONMixin):
    appid: str
    cameras: List[Camera]
    channelName: str
    token: str
    uid: int



import asyncio
import logging
from abc import abstractmethod
from typing import Any, Awaitable, Callable

import betterproto

from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model import RegionData
from pymammotion.data.model.device import MowingDevice
from pymammotion.data.state_manager import StateManager
from pymammotion.proto import has_field
from pymammotion.proto.luba_msg import LubaMsg
from pymammotion.proto.mctrl_nav import NavGetCommDataAck, NavGetHashListAck
from pymammotion.utility.movement import get_percent, transform_both_speeds

_LOGGER = logging.getLogger(__name__)


def find_next_integer(lst: list[int], current_hash: int) -> int | None:
    try:
        # Find the index of the current integer
        current_index = lst.index(current_hash)

        # Check if there is a next integer in the list
        if current_index + 1 < len(lst):
            return lst[current_index + 1]
        else:
            return None  # Or raise an exception or handle it in some other way
    except ValueError:
        # Handle the case where current_int is not in the list
        return None  # Or raise an exception or handle it in some other way


class MammotionBaseDevice:
    """Base class for Mammotion devices."""

    _mower: MowingDevice
    _state_manager: StateManager
    _cloud_device: Device | None = None

    def __init__(self, device: MowingDevice, cloud_device: Device | None = None) -> None:
        """Initialize MammotionBaseDevice."""
        self.loop = asyncio.get_event_loop()
        self._raw_data = LubaMsg().to_dict(casing=betterproto.Casing.SNAKE)
        self._mower = device
        self._state_manager = StateManager(self._mower)
        self._state_manager.gethash_ack_callback = self.datahash_response
        self._state_manager.get_commondata_ack_callback = self.commdata_response
        self._notify_future: asyncio.Future[bytes] | None = None
        self._cloud_device = cloud_device

    def set_notification_callback(self, func: Callable[[], Awaitable[None]]) -> None:
        self._state_manager.on_notification_callback = func

    def set_queue_callback(self, func: Callable[[str, dict[str, Any]], Awaitable[bytes]]) -> None:
        self._state_manager.queue_command_callback = func

    async def datahash_response(self, hash_ack: NavGetHashListAck) -> None:
        """Handle datahash responses."""
        await self.queue_command("synchronize_hash_data", hash_num=hash_ack.data_couple[0])

    async def commdata_response(self, common_data: NavGetCommDataAck) -> None:
        """Handle common data responses."""
        total_frame = common_data.total_frame
        current_frame = common_data.current_frame

        missing_frames = self._mower.map.missing_frame(common_data)
        if len(missing_frames) == 0:
            # get next in hash ack list

            data_hash = find_next_integer(self._mower.nav.toapp_gethash_ack.data_couple, common_data.hash)
            if data_hash is None:
                return

            await self.queue_command("synchronize_hash_data", hash_num=data_hash)
        else:
            if current_frame != missing_frames[0] - 1:
                current_frame = missing_frames[0] - 1

            region_data = RegionData()
            region_data.hash = common_data.hash
            region_data.action = common_data.action
            region_data.type = common_data.type
            region_data.total_frame = total_frame
            region_data.current_frame = current_frame
            await self.queue_command("get_regional_data", regional_data=region_data)

    def _update_raw_data(self, data: bytes) -> None:
        """Update raw and model data from notifications."""
        tmp_msg = LubaMsg().parse(data)
        res = betterproto.which_one_of(tmp_msg, "LubaSubMsg")
        match res[0]:
            case "nav":
                self._update_nav_data(tmp_msg)
            case "sys":
                self._update_sys_data(tmp_msg)
            case "driver":
                self._update_driver_data(tmp_msg)
            case "net":
                self._update_net_data(tmp_msg)
            case "mul":
                self._update_mul_data(tmp_msg)
            case "ota":
                self._update_ota_data(tmp_msg)

        self._mower.update_raw(self._raw_data)

    def _update_nav_data(self, tmp_msg) -> None:
        """Update navigation data."""
        nav_sub_msg = betterproto.which_one_of(tmp_msg.nav, "SubNavMsg")
        if nav_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", nav_sub_msg[0])
            return
        nav = self._raw_data.get("nav", {})
        if isinstance(nav_sub_msg[1], int):
            nav[nav_sub_msg[0]] = nav_sub_msg[1]
        else:
            nav[nav_sub_msg[0]] = nav_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["nav"] = nav

    def _update_sys_data(self, tmp_msg) -> None:
        """Update system data."""
        sys_sub_msg = betterproto.which_one_of(tmp_msg.sys, "SubSysMsg")
        if sys_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", sys_sub_msg[0])
            return
        sys = self._raw_data.get("sys", {})
        sys[sys_sub_msg[0]] = sys_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["sys"] = sys

    def _update_driver_data(self, tmp_msg) -> None:
        """Update driver data."""
        drv_sub_msg = betterproto.which_one_of(tmp_msg.driver, "SubDrvMsg")
        if drv_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", drv_sub_msg[0])
            return
        drv = self._raw_data.get("driver", {})
        drv[drv_sub_msg[0]] = drv_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["driver"] = drv

    def _update_net_data(self, tmp_msg) -> None:
        """Update network data."""
        net_sub_msg = betterproto.which_one_of(tmp_msg.net, "NetSubType")
        if net_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", net_sub_msg[0])
            return
        net = self._raw_data.get("net", {})
        if isinstance(net_sub_msg[1], int):
            net[net_sub_msg[0]] = net_sub_msg[1]
        else:
            net[net_sub_msg[0]] = net_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["net"] = net

    def _update_mul_data(self, tmp_msg) -> None:
        """Update mul data."""
        mul_sub_msg = betterproto.which_one_of(tmp_msg.mul, "SubMul")
        if mul_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", mul_sub_msg[0])
            return
        mul = self._raw_data.get("mul", {})
        mul[mul_sub_msg[0]] = mul_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["mul"] = mul

    def _update_ota_data(self, tmp_msg) -> None:
        """Update OTA data."""
        ota_sub_msg = betterproto.which_one_of(tmp_msg.ota, "SubOtaMsg")
        if ota_sub_msg[1] is None:
            _LOGGER.debug("Sub message was NoneType %s", ota_sub_msg[0])
            return
        ota = self._raw_data.get("ota", {})
        ota[ota_sub_msg[0]] = ota_sub_msg[1].to_dict(casing=betterproto.Casing.SNAKE)
        self._raw_data["ota"] = ota

    @property
    def raw_data(self) -> dict[str, Any]:
        """Get the raw data of the device."""
        return self._raw_data

    @property
    def mower(self) -> MowingDevice:
        """Get the LubaMsg of the device."""
        return self._mower

    @abstractmethod
    async def queue_command(self, key: str, **kwargs: any) -> bytes | None:
        """Queue commands to mower."""

    @abstractmethod
    async def _ble_sync(self):
        """Send ble sync command every 3 seconds or sooner."""

    @abstractmethod
    def stop(self):
        """Stop everything ready for destroying."""

    async def start_sync(self, retry: int) -> None:
        """Start synchronization with the device."""
        await self.queue_command("get_device_base_info")
        await self.queue_command("get_device_product_model")
        await self.queue_command("get_report_cfg")
        """RTK and dock location."""
        await self.queue_command("allpowerfull_rw", id=5, rw=1, context=1)

    async def start_map_sync(self) -> None:
        """Start sync of map data."""
        try:
            # work out why this crashes sometimes for better proto

            if self._cloud_device and len(self._mower.map.area_name) == 0:
                await self.queue_command("get_area_name_list", device_id=self._cloud_device.iotId)
        except Exception:
            """Do nothing for now."""

        await self.queue_command("read_plan", sub_cmd=2, plan_index=0)

        if not has_field(self.mower.nav.toapp_gethash_ack):
            await self.queue_command("get_all_boundary_hash_list", sub_cmd=0)
            await self.queue_command("get_hash_response", total_frame=1, current_frame=1)
        else:
            for data_hash in self.mower.nav.toapp_gethash_ack.data_couple:
                await self.queue_command("synchronize_hash_data", hash_num=data_hash)

        # sub_cmd 3 is job hashes??
        # sub_cmd 4 is dump location (yuka)
        # jobs list
        # hash_list_result = await self._send_command_with_args("get_all_boundary_hash_list", sub_cmd=3)

    async def async_get_errors(self) -> None:
        """Error codes."""
        await self.queue_command("allpowerfull_rw", id=5, rw=1, context=2)
        await self.queue_command("allpowerfull_rw", id=5, rw=1, context=3)

    async def move_forward(self, linear: float) -> None:
        """Move forward. values 0.0 1.0."""
        linear_percent = get_percent(abs(linear * 100))
        (linear_speed, angular_speed) = transform_both_speeds(90.0, 0.0, linear_percent, 0.0)
        await self.queue_command("send_movement", linear_speed=linear_speed, angular_speed=angular_speed)

    async def move_back(self, linear: float) -> None:
        """Move back. values 0.0 1.0."""
        linear_percent = get_percent(abs(linear * 100))
        (linear_speed, angular_speed) = transform_both_speeds(270.0, 0.0, linear_percent, 0.0)
        await self.queue_command("send_movement", linear_speed=linear_speed, angular_speed=angular_speed)

    async def move_left(self, angulur: float) -> None:
        """Move forward. values 0.0 1.0."""
        angular_percent = get_percent(abs(angulur * 100))
        (linear_speed, angular_speed) = transform_both_speeds(0.0, 0.0, 0.0, angular_percent)
        await self.queue_command("send_movement", linear_speed=linear_speed, angular_speed=angular_speed)

    async def move_right(self, angulur: float) -> None:
        """Move back. values 0.0 1.0."""
        angular_percent = get_percent(abs(angulur * 100))
        (linear_speed, angular_speed) = transform_both_speeds(0.0, 180.0, 0.0, angular_percent)
        await self.queue_command("send_movement", linear_speed=linear_speed, angular_speed=angular_speed)

    async def command(self, key: str, **kwargs):
        """Send a command to the device."""
        return await self.queue_command(key, **kwargs)
